@@ -33,8 +33,10 @@ typedef struct _NVRegistry NVRegistry;
 typedef struct _NVIndexEntry NVIndexEntry;
 typedef struct _NVEntry NVEntry;
 typedef guint32 NVHandle;
-typedef gboolean (*NVTableForeachFunc)(NVHandle handle, const gchar *name, const gchar *value, gssize value_len,
-                                       gpointer user_data);
+typedef guint8 NVType;
+typedef gboolean (*NVTableForeachFunc)(NVHandle handle, const gchar *name,
+                                       const gchar *value, gssize value_len,
+                                       NVType type, gpointer user_data);
 typedef gboolean (*NVTableForeachEntryFunc)(NVHandle handle, NVEntry *entry, NVIndexEntry *index_entry,
                                             gpointer user_data);
 
@@ -116,9 +118,6 @@ typedef struct _NVReferencedSlice
   NVHandle handle;
   guint32 ofs;
   guint32 len;
-  guint8 type;
-
-  gchar name[0];
 } NVReferencedSlice;
 
 /*
@@ -133,14 +132,32 @@ struct _NVEntry
     {
       /* make sure you don't exceed 8 bits here. So if you want to add new
        * bits, decrease the size of __bit_padding below */
+
+      /* some of these bits were not zero initialized in the past, which we
+       * are fixing by the use of the NVT_SUPPORTS_UNSET flag in the NVTable
+       * header.  If that flag is not present, we fix all but the originally
+       * existing bit fields to zero (both in current and legacy
+       * deserializers).  We are using
+       * NVENTRY_FLAGS_DEFINED_IN_LEGACY_FORMATS as a bitmask to mask out
+       * "indirect" and "referenced" in the "flags" member below, which is
+       * unioned on the bitfield.
+       */
       guint8 indirect:1,
              referenced:1,
              unset:1,
-             __bit_padding:5;
+             type_present:1,
+             __bit_padding:4;
     };
     guint8 flags;
   };
   guint8 name_len;
+  NVType type;
+
+  /* NOTE: this field fills an empty padding byte, so if you are adding
+   * fields, please remove this.  This is now zero initialized as of the
+   * first version of syslog-ng that contain this member.  Earlier versions
+   * had an uninitialized padding byte here.  */
+  guint8 __reserved;
   guint32 alloc_len;
   union
   {
@@ -148,15 +165,35 @@ struct _NVEntry
     {
       guint32 value_len;
       /* variable data, first the name of this entry, then the value, both are NUL terminated */
-      gchar data[0];
+      gchar data[];
     } vdirect;
 
-    NVReferencedSlice vindirect;
+    struct
+    {
+      NVHandle handle;
+      guint32 ofs;
+      guint32 len;
+
+      /* NOTE: this type field was promoted up into NVEntry, so we won't need
+       * the "type" field here, we are moving the value stored here in existing
+       * serialized messages in _update_entry() in logmsg-serialize-fixup.c.  We
+       * might be able to reuse this byte once we drop compatibility with
+       * version 26 of the LogMessage serialization format.
+       *
+       * NOTE: we zero out this field upon reading and then reserializing a message.
+       */
+
+      guint8 __deprecated_type_field;
+
+      gchar name[0];
+    } vindirect;
   };
 };
 
 #define NV_ENTRY_DIRECT_HDR ((gsize) (&((NVEntry *) NULL)->vdirect.data))
+#define NV_ENTRY_DIRECT_SIZE(name_len, value_len) ((value_len) + NV_ENTRY_DIRECT_HDR + (name_len) + 2)
 #define NV_ENTRY_INDIRECT_HDR (sizeof(NVEntry))
+#define NV_ENTRY_INDIRECT_SIZE(name_len) (NV_ENTRY_INDIRECT_HDR + name_len + 1)
 
 static inline const gchar *
 nv_entry_get_name(NVEntry *self)
@@ -258,22 +295,33 @@ struct _NVTable
  * static values */
 #define NV_TABLE_MIN_BYTES  128
 
-gboolean nv_table_add_value(NVTable *self, NVHandle handle, const gchar *name, gsize name_len, const gchar *value,
-                            gsize value_len, gboolean *new_entry);
-void nv_table_unset_value(NVTable *self, NVHandle handle);
-gboolean nv_table_add_value_indirect(NVTable *self, NVHandle handle, const gchar *name, gsize name_len,
-                                     NVReferencedSlice *referenced_slice, gboolean *new_entry);
+gboolean nv_table_add_value(NVTable *self, NVHandle handle,
+                            const gchar *name, gsize name_len,
+                            const gchar *value, gsize value_len,
+                            NVType type, gboolean *new_entry);
+gboolean nv_table_unset_value(NVTable *self, NVHandle handle);
+gboolean nv_table_add_value_indirect(NVTable *self, NVHandle handle,
+                                     const gchar *name, gsize name_len,
+                                     NVReferencedSlice *referenced_slice,
+                                     NVType type, gboolean *new_entry);
 
 gboolean nv_table_foreach(NVTable *self, NVRegistry *registry, NVTableForeachFunc func, gpointer user_data);
 gboolean nv_table_foreach_entry(NVTable *self, NVTableForeachEntryFunc func, gpointer user_data);
 
 NVTable *nv_table_new(gint num_static_values, gint index_size_hint, gint init_length);
 NVTable *nv_table_init_borrowed(gpointer space, gsize space_len, gint num_static_entries);
-gboolean nv_table_realloc(NVTable *self, NVTable **new);
+gboolean nv_table_realloc(NVTable *self, NVTable **new_nv_table);
+NVTable *nv_table_compact(NVTable *self);
 NVTable *nv_table_clone(NVTable *self, gint additional_space);
 NVTable *nv_table_ref(NVTable *self);
 void nv_table_unref(NVTable *self);
 
+static inline gboolean
+nv_table_is_handle_static(NVTable *self, NVHandle handle)
+{
+  return (handle <= self->num_static_entries);
+}
+
 static inline gsize
 nv_table_get_alloc_size(gint num_static_entries, gint index_size_hint, gint init_length)
 {
@@ -305,68 +353,74 @@ static inline gchar *
 nv_table_get_ofs_table_top(NVTable *self)
 {
   return (gchar *) &self->data[self->num_static_entries * sizeof(self->static_entries[0]) +
-                               self->index_size * sizeof(NVIndexEntry)];
+                                                        self->index_size * sizeof(NVIndexEntry)];
 }
 
 static inline gboolean
 nv_table_alloc_check(NVTable *self, gsize alloc_size)
 {
-  if (nv_table_get_bottom(self) - nv_table_get_ofs_table_top(self) < alloc_size)
+  if ((gsize)(nv_table_get_bottom(self) - nv_table_get_ofs_table_top(self)) < alloc_size)
     return FALSE;
   return TRUE;
 }
 
 /* private declarations for inline functions */
-NVEntry *nv_table_get_entry_slow(NVTable *self, NVHandle handle, NVIndexEntry **index_entry);
+NVEntry *nv_table_get_entry_slow(NVTable *self, NVHandle handle, NVIndexEntry **index_entry, NVIndexEntry **index_slot);
 const gchar *nv_table_resolve_indirect(NVTable *self, NVEntry *entry, gssize *len);
 
 
 static inline NVEntry *
-__nv_table_get_entry(NVTable *self, NVHandle handle, guint16 num_static_entries, NVIndexEntry **index_entry)
+__nv_table_get_entry(NVTable *self, NVHandle handle, guint16 num_static_entries, NVIndexEntry **index_entry,
+                     NVIndexEntry **index_slot)
 {
   guint32 ofs;
+  NVIndexEntry *t1, *t2;
+
+  if (!index_entry)
+    index_entry = &t1;
+  if (!index_slot)
+    index_slot = &t2;
 
   if (G_UNLIKELY(!handle))
     {
       *index_entry = NULL;
+      *index_slot = NULL;
       return NULL;
     }
 
-  if (G_LIKELY(handle <= num_static_entries))
+  if (G_LIKELY(nv_table_is_handle_static(self, handle)))
     {
       ofs = self->static_entries[handle - 1];
       *index_entry = NULL;
+      *index_slot = NULL;
       if (G_UNLIKELY(!ofs))
         return NULL;
       return (NVEntry *) (nv_table_get_top(self) - ofs);
     }
   else
     {
-      return nv_table_get_entry_slow(self, handle, index_entry);
+      return nv_table_get_entry_slow(self, handle, index_entry, index_slot);
     }
 }
 
 static inline NVEntry *
-nv_table_get_entry(NVTable *self, NVHandle handle, NVIndexEntry **index_entry)
+nv_table_get_entry(NVTable *self, NVHandle handle, NVIndexEntry **index_entry, NVIndexEntry **index_slot)
 {
-  return __nv_table_get_entry(self, handle, self->num_static_entries, index_entry);
+  return __nv_table_get_entry(self, handle, self->num_static_entries, index_entry, index_slot);
 }
 
 static inline gboolean
 nv_table_is_value_set(NVTable *self, NVHandle handle)
 {
-  NVIndexEntry *index_entry;
-
-  return nv_table_get_entry(self, handle, &index_entry) != NULL;
+  return nv_table_get_entry(self, handle, NULL, NULL) != NULL;
 }
 
 static inline const gchar *
-nv_table_get_value_if_set(NVTable *self, NVHandle handle, gssize *length)
+nv_table_get_value(NVTable *self, NVHandle handle, gssize *length, NVType *type)
 {
   NVEntry *entry;
-  NVIndexEntry *index_entry;
 
-  entry = nv_table_get_entry(self, handle, &index_entry);
+  entry = nv_table_get_entry(self, handle, NULL, NULL);
   if (!entry || entry->unset)
     {
       if (length)
@@ -374,6 +428,8 @@ nv_table_get_value_if_set(NVTable *self, NVHandle handle, gssize *length)
       return NULL;
     }
 
+  if (type)
+    *type = entry->type;
   if (!entry->indirect)
     {
       if (length)
@@ -383,16 +439,6 @@ nv_table_get_value_if_set(NVTable *self, NVHandle handle, gssize *length)
   return nv_table_resolve_indirect(self, entry, length);
 }
 
-static inline const gchar *
-nv_table_get_value(NVTable *self, NVHandle handle, gssize *length)
-{
-  const gchar *value = nv_table_get_value_if_set(self, handle, length);
-
-  if (!value)
-    return null_string;
-  return value;
-}
-
 static inline NVIndexEntry *
 nv_table_get_index(NVTable *self)
 {
